// Sample file provided by Nicolas Bonneel (http://www.bonneel.com)
// TreeGenerator sample (http://www.treegenerator.com)

#include "OgreTreeGeneratorApplication.h"
#include "OgreTreeGeneratorShadowmapListener.h"
#include "shadowmap.h"

#define NXTREES 200
#define NZTREES 200


void OgreTreeGeneratorApplication::createScene(void)
{		
	// Create a light
	Light* l = mSceneMgr->createLight("MainLight");
	l->setPosition(-230,270,-40);
	l->setDirection(Vector3(0.4,-1.7,-0.1).normalisedCopy());
	l->setCastShadows(false); // we cast shadows by hand

	// Manual shadowmap camera... fit by hand
	Camera* lightCamera = mSceneMgr->createCamera("lightCamera");
	lightCamera->setDirection(l->getDirection());
	lightCamera->setPosition(l->getPosition());
	Matrix4 M;
	mRoot->getRenderSystem()->_makeProjectionMatrix(Degree(120),1.,0.1,500,M, true); 
	lightCamera->setCustomProjectionMatrix(true, M);

	// No need for a high quality shadowmap since we won't be able to get leaf-level detailed shadows even with a 4k*4k shadowmap.
	RenderTarget* shadowmapRTT = mRoot->getRenderSystem()->createRenderTexture( "ShadowMap",512,512,TEX_TYPE_2D,PF_FLOAT16_RGBA ); 
	Viewport* shadowmapRTTVp = shadowmapRTT->addViewport(lightCamera); 
	shadowmapRTTVp->setMaterialScheme("shadowCasting");
	shadowmapRTTVp->setClearEveryFrame(true); 
	shadowmapRTTVp->setOverlaysEnabled(false); 
	shadowmapRTT->addListener(new ShadowMapTargetListener()); 

	// load the heightmap (only to place trees... not for the terrain geometry)
	TexturePtr tex1 = TextureManager::getSingleton().load("heightmap.bmp", ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);
	HardwarePixelBufferSharedPtr heightmapBuffer = tex1->getBuffer();
	PixelBox pixelBox = heightmapBuffer->lock(Image::Box(0, 0, tex1->getWidth(), tex1->getHeight()), HardwareBuffer::HBL_READ_ONLY); 
	mHeightFieldBits = static_cast<Ogre::uint8*>(pixelBox.data);

	// Create the terrain
	mTerrainEnt = mSceneMgr->createEntity("terrain", "terrain.mesh");
	mTerrainEnt->setNormaliseNormals(true);                                 // used because terrain is scaled (later)
	mTerrainEnt->setRenderQueueGroup(RENDER_QUEUE_BACKGROUND );             // to force terrain beeing rendered first (fillrate optimization since billboards' fragment programs are fillrate consuming)
	SceneNode* terrainNode = mSceneMgr->getRootSceneNode()->createChildSceneNode();
	terrainNode->attachObject(mTerrainEnt);
	terrainNode->setScale(10,10,10);

	// Create the skydome
	Entity* SkydDomeEnt = mSceneMgr->createEntity("Skydome", "skydome.mesh");
	SkydDomeEnt->setNormaliseNormals(true);
	SceneNode* skydomeNode = mSceneMgr->getRootSceneNode()->createChildSceneNode();
	skydomeNode->attachObject(SkydDomeEnt);
	skydomeNode->setScale(10,10,10);

	// Pre-load 3 trees
	Entity* tree1 = mSceneMgr->createEntity("billboard2", "tree2Billboard20.mesh");
	tree1->setVisible(false);
	tree1->setNormaliseNormals(true);
	Entity *tree2 = mSceneMgr->createEntity("billboard3", "tree3Billboard11.mesh");
	tree2->setVisible(false);
	tree2->setNormaliseNormals(true);
	Entity * tree3 = mSceneMgr->createEntity("billboard4", "tree4Billboard15.mesh");
	tree3->setVisible(false);
	tree3->setNormaliseNormals(true);

	// Cannot use StaticGeometry with manual LODs... we choose to get rid of LOD and render a constant quality billboard cloud forest
	SceneNode* treesNode = mSceneMgr->getRootSceneNode()->createChildSceneNode("treesNode");
	for (int i=-NXTREES/2; i<NXTREES/2; i++)
		for (int j=-NZTREES/2; j<NZTREES/2; j++)
		{
			Entity* cloneEnt;
			int seed = rand()%1000; // take a random tree
			if (seed<400)			
				cloneEnt = tree1->clone("billboardClone"+StringConverter::toString((i+NXTREES/2)*NZTREES+j));
			else
				if (seed<700)	
					cloneEnt = tree2->clone("billboardClone"+StringConverter::toString((i+NXTREES/2)*NZTREES+j));
				else
					cloneEnt = tree3->clone("billboardClone"+StringConverter::toString((i+NXTREES/2)*NZTREES+j));

			cloneEnt->setNormaliseNormals(true);
			SceneNode* node = treesNode->createChildSceneNode();
			node->attachObject(cloneEnt);
			float x = (i+(rand()%1)*0.06)*1000./NXTREES; // jitter around a uniform grid
			float z = (j+(rand()%1)*0.06)*1000./NZTREES;
			float y = getTerrainHeight(Vector3(x,0,z));

			float s = 0.1*(0.6 + (rand()%100)*0.01*0.6); // scale randomly too
			node->setPosition(x,y,z);
			node->setScale(s,s,s);
			node->setOrientation(Quaternion(Degree((rand()%3600)*0.1),Vector3::UNIT_Y)); // rotate randomly
		}


		StaticGeometry* staticForest = mSceneMgr->createStaticGeometry("staticForest"); //create the static geometry based on the scenenode
		staticForest->addSceneNode(treesNode);
		staticForest->build();
		mSceneMgr->getRootSceneNode()->removeChild(treesNode); // we remove the scenenode we don't use anymore

		heightmapBuffer->unlock(); // unlock heightfield, we don't need it anymore

		mCamera->setPosition(-23,88,213);

		// just add the shadow textureunit in the following materials
		mShadowReceivers.push_back(((MaterialPtr)MaterialManager::getSingleton().getByName("lambert3"))->getTechnique(0)->getPass(0));
		mShadowReceivers.push_back(((MaterialPtr)MaterialManager::getSingleton().getByName("matBillboard12A"))->getTechnique(0)->getPass(0));
		mShadowReceivers.push_back(((MaterialPtr)MaterialManager::getSingleton().getByName("tree3billboard11Mat"))->getTechnique(0)->getPass(0));
		mShadowReceivers.push_back(((MaterialPtr)MaterialManager::getSingleton().getByName("tree4Billboard15Mat"))->getTechnique(0)->getPass(0));

		for (int i=0; i<mShadowReceivers.size(); i++)
			mShadowReceivers[i]->createTextureUnitState("ShadowMap");

}

float OgreTreeGeneratorApplication::getTerrainHeight(Vector3 pos)
{
	int x = pos.z*1024/1000.+512;
	int y = pos.x*1024/1000.+512;
	return mHeightFieldBits[x*1024+y]*0.37;
}


// Create new frame listener
void OgreTreeGeneratorApplication::createFrameListener(void)
{
	mFrameListener= new OgreTreeGeneratorFrameListener(mWindow, mCamera, mSceneMgr, &mShadowReceivers);
	mRoot->addFrameListener(mFrameListener);
}


bool OgreTreeGeneratorApplication::setup()
{
	ExampleApplication::setup();
	LogManager::getSingleton().setLogDetail( LL_BOREME );
	return true;
}

